/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   decomp.c
 *
 *  PURPOSE:
 *	  contains free-standing video bunch decompression code
 *
 *  FUNCTIONS:
 *	  DecompBunch(<Bunch data>)
 *
 *  COMMENTS:
 *    In this case a video bunch means a collection of frames which are all to
 *    be decompressed together, i.e. a key frame and all the frames up to but
 *    not including the next key frame.
 *
 *    This is part of the distributed recompression tool, so typically
 *    compression takes place after decompression.  This means the bunch size
 *    is usually already determined, by the occurrence of key frames in the
 *    original compressed source material.
 *
 *    '-In-' refers to input (uncompressed) data
 *    '-Out-' refers to output (recompressed) data
 */

#include	"dcomp.h"
#include	"dcomp_s.h"


//
//  FUNCTION: DecompBunch
//
//  PURPOSE:
//    decompress the frames in a bunch using ICM calls
//
//  PARAMETERS:
//    lBunchId      - unique bunch ID - frame number of first frame in bunch
//    lBunchLen     - number of frames in the bunch
//    lpbiIn        - pointer to video format for input data
//    lpbiOut       - pointer to video format for output data
//    alIn[]        - array of indexes of frame starts in input video buffer
//    lplInDataLen  - total length of the input video data
//    lpbInData[]   - pointer to the input video data buffer
//    lpbOutData[]  - pointer to buffer to receive the output video data
//
//  RETURN VALUE:
//    Zero on success, non-zero on error
//
//  COMMENTS:
//    Decompression of every frame is done into the same buffer, since some
//    decompressors rely on this.  Each frame is then moved to is own buffer
//    before decompressing the next frame.
//

extern HINSTANCE ghInstance;


extern HWND hWnd;


LONG		DecompBunch
(
	LONG			lBunchId,
	LONG			lBunchLen,
	LPBITMAPINFO	lpbiIn,
	LPBITMAPINFO	lpbiOut,
	LONG			alIn[],
	LONG			lInDataLen,
	BYTE			lpbInData[],
	BYTE			lpbOutData[]
)
{
	HIC					hIC;
	LONG				l;
    HRESULT				hr;
	DWORD				dwFlags;
	LONG				lBytes[MAXBUNCHLEN];
	LPBITMAPINFOHEADER	lpbihIn = &lpbiIn->bmiHeader;
	LPBITMAPINFOHEADER	lpbihOut = &lpbiOut->bmiHeader;

	LPVOID				lpBuildFrame;


	
	// Compute the length of each frame

	for (l = 0; l < lBunchLen - 1; l++)
	{
		lBytes[l] = alIn[l + 1] - alIn[l];
	}

	lBytes[l] = lInDataLen - alIn[l];		// l is now (lBunchLen - 1)


	// Open decompressor

	if ((hIC = ICLocate(ICTYPE_VIDEO, lpbihIn->biCompression,
									lpbihIn, NULL, ICMODE_DECOMPRESS)) != NULL)
	{
		// Initialise compressor

		Eprintf(TEXT("%s: Decompressing bunch %ld\n"),
										_tgetenv(TEXT("COMPUTERNAME")), lBunchId);

		if (ICDecompressBegin(hIC, lpbiIn, lpbiOut) == ICERR_OK)
		{
			// Designate last buffer as working buffer for the decompressor

			lpBuildFrame = &lpbOutData[lpbihOut->biSizeImage * (lBunchLen - 1)];


			for (l = 0L; l < lBunchLen; l++)
			{
			    // Decompress the data

				dwFlags = 	(l == 0 ? 0UL : ICDECOMPRESS_NOTKEYFRAME)
						  |
						  	(lBytes[l] == 0 ? ICDECOMPRESS_NULLFRAME : 0UL)
						  ;
								// N.B. assumes only one key frame in bunch
								// - need a better solution


				// Should use ICDecompressEx etc. - need lots more arguments.

		        if (ICDecompress(	hIC, dwFlags,
			        				lpbihIn, &lpbInData[alIn[l]],
			        				lpbihOut,
			        				lpBuildFrame
			        			) != ICERR_OK)
			    {
					Eprintf(TEXT("Problem decompressing frame %ld of bunch\n"), l);
					if ((hr = ICClose(hIC)) != ICERR_OK)
					{
						Eprintf(
							TEXT("Error code %ld closing decompressor ")
							TEXT("\"%4.4hs\"\n"),
										hr, &lpbihIn->biCompression);
					}
					return(1L);
		        }
				

				// Need to copy result to the right place if not last frame.
				// If a null frame, should copy the previous frame instead
				// - this is because decompressors apparently can't be trusted
				// to leave the output buffer alone when the frame is null.

				if (dwFlags & ICDECOMPRESS_NULLFRAME)
				{
					if (l > 0L)		// better hope first frame isn't null
					{
						memcpy
						(
							&lpbOutData[lpbihOut->biSizeImage * l],
							&lpbOutData[lpbihOut->biSizeImage * (l - 1)],
							lpbihOut->biSizeImage
						);
					}
				}
				else
				{
					if (l < (lBunchLen - 1))
					{
						memcpy(&lpbOutData[lpbihOut->biSizeImage * l],
										lpBuildFrame, lpbihOut->biSizeImage);
					}
				}
			

                // DIB display code
                if(IsWindow(hWnd))
                {
                    HDC hdc;
                    //PGETFRAME pGetFrame;
                    HDRAWDIB hDrawDib;
                    //LPBITMAPINFOHEADER lpbi;
                        

                    // Open the DIB draw routines
                    hdc = GetDC(hWnd);
                    hDrawDib = DrawDibOpen();
                    
                    // Size the window and draw
                    SetWindowPos(hWnd,HWND_TOP,0,0,lpbihOut->biWidth, lpbihOut->biHeight,SWP_NOMOVE);

                    DrawDibDraw(hDrawDib,hdc, 0, 0, lpbihOut->biWidth, lpbihOut->biHeight,
  			        				   lpbihOut,lpBuildFrame,
								       0, 0, -1, -1,
    							       0 );

                    // Close the DIB draw routines
                    DrawDibClose(hDrawDib);
                    ReleaseDC(hWnd,hdc);
                }

			}
 
		    if ((hr = ICDecompressEnd(hIC)) != ICERR_OK)
			{
				Eprintf(TEXT("Error code %ld (0x%lx) ending decompression\n"),
																	hr, hr);
				if ((hr = ICClose(hIC)) != ICERR_OK)
				{
					Eprintf(
						TEXT("Error code %ld closing decompressor \"%4.4hs\"\n"),
										hr, &lpbihIn->biCompression);
				}
				return(1L);
			}

		}
		else
		{
			Eprintf(TEXT("Couldn't initialise decompressor \"%4.4hs\"\n"),
											&lpbihIn->biCompression);
			if ((hr = ICClose(hIC)) != ICERR_OK)
			{
				Eprintf(TEXT("Error code %ld closing decompressor \"%4.4hs\"\n"),
										hr, &lpbihIn->biCompression);
			}
			return(1L);
		}
	}
	else
	{
		Eprintf(TEXT("Couldn't open decompressor \"%4.4hs\"\n"),
											&lpbihIn->biCompression);
		return(1L);
	}

	if ((hr = ICClose(hIC)) != ICERR_OK)
	{
		Eprintf(TEXT("Error code %ld closing decompressor \"%4.4hs\"\n"), hr,
											&lpbihIn->biCompression);
		return(1L);
	}

//	DestroyWindow(hWnd);
	
	return(0L);
}
