/* THIS CODE AND INFORMATION IS PROVIDED "AS IS" WITHOUT WARRANTY OF
 * ANY KIND, EITHER EXPRESSED OR IMPLIED, INCLUDING BUT NOT LIMITED TO
 * THE IMPLIED WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A
 * PARTICULAR PURPOSE.
 *
 * Copyright (C) 1994, 1995  Microsoft Corporation.  All Rights Reserved.
 *
 *  MODULE:   decompq.c
 *
 *  PURPOSE:
 *	  handles input ('-Com-') and output ('-Unc-') queues for the DecompBunch
 *    function
 *
 *  FUNCTIONS:
 *	  RPCDecompBunchRequest(<bunch_data>)
 *	  RPCGetDecompBunchResult(<bunch_data>)
 *	  RPCDecompBunchReqAccept(handle_t hHandle)
 *	  RPCDecompBunchRetAvail(handle_t hHandle)
 *	  InitDecompQueues(void)
 *	  DispDecompBunchOutQ(TCHAR *szPrefix)
 *	  DispDecompBunchInQ(TCHAR *szPrefix)
 *
 *  COMMENTS:
 *    '-Rec-' data refers to the final recompressed output: if this information
 *    is supplied, it is just passed onward towards the compressor.
 *    The functions beginning with 'RPC' are client-server interfaces defined
 *    in the MIDL file.
 */

#include	"dcomp.h"
#include	"dcomp_s.h"

void		DispDecompBunchOutQ(TCHAR *szPrefix);
void		DispDecompBunchInQ(TCHAR *szPrefix);


//
//  FUNCTION: DecompBunchPass
//
//  PURPOSE:
//    decompress a bunch from the input queue and place it on the output queue
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//    The decompression component of a single pass of CodecLoop.  Returns
//    immediately if there are no bunches to process.  If an attempt to
//    decompress a bunch fails, the bunch is re-queued for a re-try on the
//    next pass.
//

void		DecompBunchPass(void)
{
	LONG					nRequ;
	LONG					lBunchLen;
	LPBITMAPINFOHEADER		lpbihUnc;
	BUNCHACTION				baAction;
    LPBUNCHQUEUE            lpbqReturn;


	// Check there is a return slot free

	switch(baAction
		= bqDecompBunchInQ.abnchPool[bqDecompBunchInQ.nQHead].baAction)
	{
		case DECOMPRESS:
            lpbqReturn = &bqDecompBunchOutQ;
			break;

		case RECOMPRESS:
            lpbqReturn = &bqCompBunchInQ;
			break;
	}

	if (ReserveSlot(lpbqReturn) == DCOMP_QERR)
	{
		return;
	}


	// Take first request off queue

	nRequ = GrabFirstBunch(&bqDecompBunchInQ);


	// If there was no request available, some other thread beat us to it

	if (nRequ == DCOMP_QEND)
	{
		UnreserveSlot(lpbqReturn);
		return;
	}

	DispDecompBunchInQ(TEXT("DPS0: "));


	// allocate output buffer

	lBunchLen = bqDecompBunchInQ.abnchPool[nRequ].lBunchLen;
	lpbihUnc = &bqDecompBunchInQ.abnchPool[nRequ].lpbiUnc->bmiHeader;

	if
	(
		(bqDecompBunchInQ.abnchPool[nRequ].lpbUncData
			= midl_user_allocate(lpbihUnc->biSizeImage * lBunchLen))
																== NULL
	)
	{
		Eprintf(TEXT("DPS3: Couldn't allocate memory - will retry request\n"));
		goto RetryRequest;
	}

	bqDecompBunchInQ.abnchPool[nRequ].lUncDataLen
										= lpbihUnc->biSizeImage * lBunchLen;


	// Do the actual decompression

	Dprintf(TEXT("DPS4: Decompressing bunch %d\n"),
								bqDecompBunchInQ.abnchPool[nRequ].lBunchId);

	if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_IDLE))
	{
		Eprintf(TEXT("Failed to set CodecLoop thread priority below normal")
											TEXT(" before decompressing\n"));
	}

	bqDecompBunchInQ.abnchPool[nRequ].lReturnStatus =
	DecompBunch
	(
		bqDecompBunchInQ.abnchPool[nRequ].lBunchId,
		bqDecompBunchInQ.abnchPool[nRequ].lBunchLen,
		bqDecompBunchInQ.abnchPool[nRequ].lpbiCom,
		bqDecompBunchInQ.abnchPool[nRequ].lpbiUnc,
		bqDecompBunchInQ.abnchPool[nRequ].lplComIndex,
		bqDecompBunchInQ.abnchPool[nRequ].lComDataLen,
		bqDecompBunchInQ.abnchPool[nRequ].lpbComData,
		bqDecompBunchInQ.abnchPool[nRequ].lpbUncData
	);

	if (!SetThreadPriority(GetCurrentThread(), THREAD_PRIORITY_NORMAL))
	{
		Eprintf(TEXT("Failed to set CodecLoop thread priority normal")
											TEXT(" after decompressing\n"));
	}


	// Add result to decompression output queue or compression
	// input queue according to the value of baAction

	EnterCriticalSection(&lpbqReturn->csQ);
	if (QueueBunch(lpbqReturn, nRequ) == DCOMP_QERR)
	{
		Eprintf(TEXT("DPS4b: Couldn't requeue bunch %ld after")
                TEXT(" decompression - will retry\n"),
                bqDecompBunchInQ.abnchPool[nRequ].lBunchId);
		LeaveCriticalSection(&lpbqReturn->csQ);
		goto RetryRequest;
	}


	// Clear away request buffers no longer needed

	ClearDown(&bqDecompBunchInQ.abnchPool[nRequ].lpbiCom);
	ClearDown(&bqDecompBunchInQ.abnchPool[nRequ].lplComIndex);
	ClearDown(&bqDecompBunchInQ.abnchPool[nRequ].lpbComData);
	LeaveCriticalSection(&lpbqReturn->csQ);

	Eprintf(
		TEXT("DPS4a: Queued decompressed bunch %d in slot %ld")
		TEXT(" after decompression\n"),
		bqDecompBunchInQ.abnchPool[nRequ].lBunchId, nRequ);


	// all OK: clear away the finished request and return

	UnreserveSlot(&bqDecompBunchInQ);
	return;

	// end of normal processing



	// error processing decompression request: cancel the request and re-queue
	// it on the input queue

RetryRequest:
	ClearDown(&bqDecompBunchInQ.abnchPool[nRequ].lpbUncData);
	QueueBunchAtHead(&bqDecompBunchInQ, nRequ);
	UnreserveSlot(lpbqReturn);
}


//
//  FUNCTION: RPCDecompBunchRequest
//
//  PURPOSE:
//    place a bunch coming from the client on the input queue
//
//  PARAMETERS:
//    hHandle           - RPC binding handle - unreferenced in this function
//    lBunchId          - bunch ID (equal to first video frame number in bunch)
//    baAction          - action to perform on bunch: DECOMPRESS or RECOMPRESS
//    lBunchLen         - number of video frames in bunch
//    lComFormatLen     - video format size for compressed data  
//    lpbComFormat[]    - video format for compressed data 
//    lUncFormatLen     - video format size for uncompressed data  
//    lpbUncFormat[]    - video format for uncompressed data 
//    lRecFormatLen     - video format size for recompressed data 
//    lpbRecFormat[]    - video format for recompressed data 
//    alCom[]           - array of indexes to compressed video frames in buffer
//    lComDataLen       - total length of the compressed video data
//    lpbComData[]      - pointer to buffer containing compressed video data
//
//  RETURN VALUE:
//    Non-zero on failure
//
//  COMMENTS:
//

LONG RPCDecompBunchRequest( 
	handle_t			hHandle,
	LONG				lBunchId,
	BUNCHACTION			baAction,
	LONG				lBunchLen,
	LONG				lComFormatLen,
	BYTE				lpbComFormat[],
	LONG				lUncFormatLen,
	BYTE				lpbUncFormat[],
	LONG				lRecFormatLen,
	BYTE				lpbRecFormat[],
	LONG				alCom[],
	LONG				lComDataLen,
	BYTE				lpbComData[]
)
{
	LONG		nRequ;

	UNREF_PARAM(hHandle);
	Dprintf(TEXT("DBQ0: Request for bunch %d received\n"), lBunchId);


	if (gfInitPending || gfServerComputerBusy)
	{
		// waiting for queues to be cleared, or computer too busy to accept
		// RPCs - return immediately with failure code

		return(DCOMP_QERR);
	}

	if (ReserveSlot(&bqDecompBunchInQ) == DCOMP_QERR)
	{
		// request queue is full - return immediately with failure code

		return(DCOMP_QERR);
	}

	if ((nRequ = GrabEmptyBunch(&bqDecompBunchInQ)) == DCOMP_QEND)
	{
		// failed - return failure code

		UnreserveSlot(&bqDecompBunchInQ);
		return(DCOMP_QERR);
	}

	Eprintf(TEXT("DBQ1: Request %d placed in slot %d\n"), lBunchId, nRequ);


	// Allocate buffers and copy parameter values into request structure.
	// Better to use persistent stub-allocated buffers without copying ...

	if
	(
			((bqDecompBunchInQ.abnchPool[nRequ].lpbiCom
								= midl_user_allocate(lComFormatLen)) == NULL)
		||	((bqDecompBunchInQ.abnchPool[nRequ].lpbiUnc
								= midl_user_allocate(lUncFormatLen)) == NULL)
		||	((bqDecompBunchInQ.abnchPool[nRequ].lpbiRec
								= midl_user_allocate(lRecFormatLen)) == NULL)
		||	((bqDecompBunchInQ.abnchPool[nRequ].lplComIndex
					= midl_user_allocate(lBunchLen * sizeof(LONG)) ) == NULL)
		||	((bqDecompBunchInQ.abnchPool[nRequ].lpbComData
								= midl_user_allocate(lComDataLen)  ) == NULL)
	)
	{
		Eprintf(TEXT("DBQ3: Couldn't allocate memory for bunch %ld")
									TEXT(" - request cancelled\n"), lBunchId);
		ClearDownBunch(&bqDecompBunchInQ, nRequ);
		UnreserveSlot(&bqDecompBunchInQ);
		return(DCOMP_QERR);
	}

	bqDecompBunchInQ.abnchPool[nRequ].lBunchId  = lBunchId;
	bqDecompBunchInQ.abnchPool[nRequ].baAction  = baAction;
	bqDecompBunchInQ.abnchPool[nRequ].lBunchLen = lBunchLen;
	bqDecompBunchInQ.abnchPool[nRequ].lComDataLen = lComDataLen;

	memcpy(bqDecompBunchInQ.abnchPool[nRequ].lpbiCom, lpbComFormat,
																lComFormatLen);
	memcpy(bqDecompBunchInQ.abnchPool[nRequ].lpbiUnc, lpbUncFormat,
																lUncFormatLen);
	memcpy(bqDecompBunchInQ.abnchPool[nRequ].lpbiRec, lpbRecFormat,
																lRecFormatLen);
	memcpy(bqDecompBunchInQ.abnchPool[nRequ].lplComIndex, alCom,
													lBunchLen * sizeof(LONG));
	memcpy(bqDecompBunchInQ.abnchPool[nRequ].lpbComData, lpbComData,
																lComDataLen);


	// add this request to the queue and return success code

	QueueBunch(&bqDecompBunchInQ, nRequ);
	return(0);
}


//
//  FUNCTION: RPCGetDecompBunchResult
//
//  PURPOSE:
//    return a decompressed bunch to the client
//
//  PARAMETERS:
//    hHandle           - RPC binding handle - unreferenced in this function
//    lplReturnStatus   - pointer to var. to receive DecompBunch return status
//    lBunchId          - pointer to variable to receive the bunch ID of the
//                        returned bunch
//    lBunchLen         - pointer to variable to receive no. of frames in bunch
//    lUncFormatLen     - video format size for uncompressed data  
//    lpbUncFormat[]    - pointer to buf. to receive uncompressed video format
//    lMaxUncDataSize   - maximum total length of the uncompressed video data
//    lplUncDataLen     - pointer to variable to receive total length of the
//                        uncompressed video data
//    lpbUncData[]      - pointer to buffer to receive uncompressed video data
//
//  RETURN VALUE:
//    Non-zero on failure
//
//  COMMENTS:
//

LONG RPCGetDecompBunchResult( 
	handle_t			hHandle,
	LONG				*lplReturnStatus,
	LONG				*lplBunchId,
	LONG				*lplBunchLen,
	LONG				lUncFormatLen,
	BYTE				lpbUncFormat[],
	LONG				lMaxUncDataSize,
	LONG				*lplUncDataLen,
	BYTE				lpbUncData[]
)
{
	LONG		nRet;

	UNREF_PARAM(hHandle);
	UNREF_PARAM(lMaxUncDataSize);

	if (gfInitPending)
	{
		// waiting for queues to be cleared - return immediately
		// with failure code

		return(DCOMP_QERR);
	}

	// Take first return (if any) off queue

	if ((nRet = GrabFirstBunch(&bqDecompBunchOutQ)) == DCOMP_QEND)
	{
		return(DCOMP_QERR);	// no return data to send so return failure code
	}

	Eprintf(TEXT("DBR1: Returning bunch %d from slot %d\n"),
			bqDecompBunchOutQ.abnchPool[nRet].lBunchId,
			nRet
		   );


	// copy return values and buffers to output arrays

	*lplReturnStatus = bqDecompBunchOutQ.abnchPool[nRet].lReturnStatus;
	*lplBunchId = bqDecompBunchOutQ.abnchPool[nRet].lBunchId;
	*lplBunchLen = bqDecompBunchOutQ.abnchPool[nRet].lBunchLen;
										// could check against lMaxBunchLen

	memcpy(lpbUncFormat,bqDecompBunchOutQ.abnchPool[nRet].lpbiUnc,
																lUncFormatLen);

	*lplUncDataLen = bqDecompBunchOutQ.abnchPool[nRet].lUncDataLen;
									// could check against lMaxUncDataSize
	memcpy(lpbUncData, bqDecompBunchOutQ.abnchPool[nRet].lpbUncData, *lplUncDataLen);


	// clear away the bunch structure and return success code

	DispDecompBunchOutQ(TEXT("DBR3: "));
	Eprintf(TEXT("Bunch %d returned after decompression\n"),
										bqDecompBunchOutQ.abnchPool[nRet].lBunchId);

	ClearDownBunch(&bqDecompBunchOutQ, nRet);
	UnreserveSlot(&bqDecompBunchOutQ); 
	return(0);
}


//
//  FUNCTION: RPCDecompBunchReqAccept
//
//  PURPOSE:
//    checks if a decompression request can be accepted, i.e. the
//    decompressor's input queue has space
//
//  PARAMETERS:
//    hHandle - RPC binding handle - unreferenced in this function
//
//  RETURN VALUE:
//    Non-zero if a bunch can be accepted
//
//  COMMENTS:
//

LONG RPCDecompBunchReqAccept(handle_t hHandle)
{
	LONG		RetVal = (bqDecompBunchInQ.nQCnt < bqDecompBunchInQ.nQLen);

	UNREF_PARAM(hHandle);

	if (gfInitPending || gfServerComputerBusy)
	{
		// waiting for queues to be cleared, or computer too busy to accept
		// RPCs - return immediately with rejection code

		return(FALSE);
	}

	Dprintf(TEXT("DAC: Request acceptance query: requests %sallowed\n"),
													(RetVal ? TEXT("") : TEXT("not ")));
	return(RetVal);
}


//
//  FUNCTION: RPCDecompBunchRetAvail
//
//  PURPOSE:
//    checks to see if checks to see if any bunches have been decompressed and
//    are available to be returned
//
//  PARAMETERS:
//    hHandle  - RPC binding handle - unreferenced in this function
//
//  RETURN VALUE:
//    Non-zero if there are any completed bunches available.
//
//  COMMENTS:
//

LONG RPCDecompBunchRetAvail(handle_t hHandle)
{
	LONG		RetVal = (bqDecompBunchOutQ.nQHead != DCOMP_QEND);

	UNREF_PARAM(hHandle);

	if (gfInitPending)
	{
		// waiting for queues to be cleared - return immediately
		// with rejection code

		return(FALSE);
	}

	Dprintf(TEXT("DAV: Availability query: %sdecompressed bunches available\n"),
								(RetVal ? TEXT("") : TEXT("no ")));
	return(RetVal);
}


//
//  FUNCTION: InitDecompQueues
//
//  PURPOSE:
//    clear down the decompressor's input and output queues to the empty
//    condition
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void		InitDecompQueues(void)
{
	LONG		i;

	EnterCriticalSection(&bqDecompBunchInQ.csQ);
	EnterCriticalSection(&bqDecompBunchOutQ.csQ);
	EnterCriticalSection(&csQPool);

	bqDecompBunchInQ.nQHead = DCOMP_QEND;
	bqDecompBunchInQ.nQCnt = 0;
	bqDecompBunchOutQ.nQHead = DCOMP_QEND;
	bqDecompBunchOutQ.nQCnt = 0;

	for (i = 0; i < BUNCHPOOLLEN; i++)
	{
		abnchQPool[i].nNext = DCOMP_QEND;
		abnchQPool[i].nStat = DCOMP_QEMPTY;

		ClearDown(&abnchQPool[i].lpbiCom);
		ClearDown(&abnchQPool[i].lpbiUnc);
		ClearDown(&abnchQPool[i].lpbiRec);
		ClearDown(&abnchQPool[i].lplComIndex);
		ClearDown(&abnchQPool[i].lpadwRetFlags);
		ClearDown(&abnchQPool[i].lpbComData);
		ClearDown(&abnchQPool[i].lpbUncData);
	}

	LeaveCriticalSection(&csQPool);
	LeaveCriticalSection(&bqDecompBunchInQ.csQ);
	LeaveCriticalSection(&bqDecompBunchOutQ.csQ);
}


//
//  FUNCTION: DispDecompBunchOutQ and DispDecompBunchInQ
//
//  PURPOSE:
//    display routines for debugging only
//
//  PARAMETERS:
//
//  RETURN VALUE:
//
//  COMMENTS:
//

void		DispDecompBunchOutQ(TCHAR *szPrefix)
{
	LONG	i, nBufPos;
	TCHAR	buf[256];

	nBufPos = _stprintf(buf, TEXT(" RET C%-2u H%-2d Q"), bqDecompBunchOutQ.nQCnt, bqDecompBunchOutQ.nQHead);
	
	if (bqDecompBunchOutQ.nQHead != DCOMP_QEND)
	{
		nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), bqDecompBunchOutQ.nQHead);

		i = bqDecompBunchOutQ.nQHead;
		while (bqDecompBunchOutQ.abnchPool[i].nNext != DCOMP_QEND)
		{
			i = bqDecompBunchOutQ.abnchPool[i].nNext;
			nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), i);
		}
	}

	Dprintf(TEXT("%sID%08.8lx%s\n"), szPrefix, GetCurrentThreadId(), buf);
}


void		DispDecompBunchInQ(TCHAR *szPrefix)
{
	LONG	i, nBufPos;
	TCHAR	buf[256];

	nBufPos = _stprintf(buf, TEXT(" REQ C%-2u H%-2d Q"), bqDecompBunchInQ.nQCnt, bqDecompBunchInQ.nQHead);

	if (bqDecompBunchInQ.nQHead != DCOMP_QEND)
	{
		nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), bqDecompBunchInQ.nQHead);

		i = bqDecompBunchInQ.nQHead;
		while (bqDecompBunchInQ.abnchPool[i].nNext != DCOMP_QEND)
		{
			i = bqDecompBunchInQ.abnchPool[i].nNext;
			nBufPos += _stprintf(&buf[nBufPos], TEXT("%d"), i);
		}
	}
	
	Dprintf(TEXT("%sID%08.8lx%s\n"), szPrefix, GetCurrentThreadId(), buf);
}
